#include <octave/oct.h>
#include <octave/parse.h>
#include <octave/f77-fcn.h>

/* 

An octave gateway routine to the FORTRAN code ode.f.  Described in 

Computer Solution of Ordinary Differential Equations: 
The Initial Value Problem  
by L. F. Shampine and M. K. Gordon.

The code can be found at

http://cm.bell-labs.com/netlib/ode/

Written by:
-- 
John L. Weatherwax                2005-10-29

email: wax@alum.mit.edu

Please send comments and especially bug reports to the
above email address.

-----
*/ 

typedef F77_RET_T (*fort_fcn_ptr_dec) (const double&,double*, double*);

extern "C"
{
  F77_RET_T F77_FUNC(ode,ODE)(fort_fcn_ptr_dec, const int&, double*, 
			      double&, double&, 
			      const double&, const double&, 
			      int&, double*, int* );
}

static octave_function *odeg_oct_fcn;
static int NEQN; 

static F77_RET_T
odeg_f_fcn(const double& t, double* x, double* xp)
{
  RowVector rv_x(NEQN); 
  for( int ii=0; ii < NEQN; ii++ ) rv_x(ii)=x[ii];

  octave_value_list args;
  args(0) = t; args(1) = rv_x; 
  int nargout = 1;

  octave_value_list retval = feval(odeg_oct_fcn, args, nargout);

  if (! error_state && retval.length () == 1)
    {
      Matrix m_xp = retval(0).matrix_value();
      for (int ii = 0; ii < NEQN; ii++){
	xp[ii] = m_xp(ii);
      }

      if (error_state)
	error ("failure evaluating user-supplied function");
    }
  else
    error ("failure evaluating user-supplied function");

  F77_RETURN(0);
}

// C++ octave API calls of the user's input function:
//
ColumnVector
ode_f_function(double t, const ColumnVector& x)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x;

  if (odeg_oct_fcn)
    {
      octave_value_list tmp = odeg_oct_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("odeg_oct_fcn");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("odeg_oct_fcn");
	}
      else
	gripe_user_supplied_eval ("odeg_oct_fcn");
    }

  return retval;
}


DEFUN_DLD (odeg, args, nargout,
  "x_t = odeg(fcn, tspan, x_0): Integrates dx/dt = fcn from tspan(1)\n\
   to tspan(end).  The array tspan can be either a two element in size\n\
   or have multiple elments, either way the output x_t represents the\n\
   solution integrated to each element in tspan")
{
  octave_value_list retval;

  if( args.length() != 3 ){ print_usage("odeg"); return retval; }
  if( nargout > 1 ) { print_usage("odeg"); return retval; }

  // Extract the user input:
  //
  odeg_oct_fcn = extract_function(args(0), "odeg", "__odeg_oct_fcn__",
	     "function F = __odeg_oct_fcn__ (t, x) ", "; endfunction");
  if(error_state){
    error( "odeg: expecting a function for the first argument" ); 
  } 
  ColumnVector TSPAN( args(1).vector_value() );
  if(error_state){
    error("odeg: expecting a vector of output times for the second argument" ); 
  }
  int nouts = TSPAN.length();
  ColumnVector x_0( args(2).vector_value() );
  if(error_state){
    error("odeg: expecting an initial condition vector for the third argument" ); 
  }
  NEQN = x_0.length();

  // FORTRAN variables:
  double RELERR = 1.e-3;
  double ABSERR = 1.e-6;
  //  double RELERR = 1.e-9;
  //  double ABSERR = 1.e-6;
  int IFLAG = +1; 

  // Check some inputs:
  if( ABSERR<=0 ){
    octave_stdout << "ABSERR must be greater than zero" << "\n"; 
    return retval; 
  }

  // Create the work/iwork vectors: 
  int NWORK = 100+21*NEQN;
  OCTAVE_LOCAL_BUFFER( double, WORKP, NWORK ); 
  int NIWORK = 5;
  OCTAVE_LOCAL_BUFFER( int, IWORKP, NIWORK );
  OCTAVE_LOCAL_BUFFER( double, XP, NEQN ); 
  for( int ii=0; ii < NEQN; ii++ ) XP[ii]=x_0(ii);

  // Create octave output matrix: 
  Matrix y_out(NEQN,nouts);
  for( int ii=0; ii < NEQN; ii++ ) y_out(ii,0)=x_0(ii);

  for( int iit=1; iit < nouts; iit++ ){ // Perform each integration step:
    double T=TSPAN(iit-1);
    double TOUT = TSPAN(iit);
    F77_FUNC(ode,ODE)(odeg_f_fcn,NEQN,XP,T,TOUT,\
		      RELERR,ABSERR,IFLAG,WORKP,IWORKP);
    if( IFLAG == 6 ){ // TODO: figure out how to send this parameter (iflag) out of this routine...
      octave_stdout << "iflag = 6 -- invalid input parameters, can't continue integration" << "\n"; 
      break; 
    }

    ColumnVector x_out(NEQN); 
    for( int ii=0; ii < NEQN; ii++ ){
      y_out(ii,iit) = XP[ii];
    }
  }

  retval.append(y_out); 
  return retval;
}
